@@ -1,9 +1,8 @@
 use std::{collections::BTreeMap, os::raw::c_char, ptr, str::FromStr, sync::Arc};
 
-use async_lock::RwLock;
+use async_lock::{Mutex as TryMutex, MutexGuardArc as TryMutexGuard, RwLock};
 use ffi_support::{rust_string_to_c, ByteBuffer, FfiStr};
 use once_cell::sync::Lazy;
-use option_lock::{Mutex as TryMutex, MutexGuardArc as TryMutexGuard};
 
 use super::{
     error::set_last_error,
@@ -84,17 +83,12 @@ where
         handle
     }
 
-    pub async fn remove(&self, handle: K) -> Result<V, Error> {
-        Arc::try_unwrap(
-            self.map
-                .write()
-                .await
-                .remove(&handle)
-                .ok_or_else(|| err_msg!("Invalid resource handle"))?
-                .1,
-        )
-        .map(|item| item.into_inner().unwrap())
-        .map_err(|_| err_msg!(Busy, "Resource handle in use"))
+    pub async fn remove(&self, handle: K) -> Option<Result<V, Error>> {
+        self.map.write().await.remove(&handle).map(|(_s, v)| {
+            Arc::try_unwrap(v)
+                .map(|item| item.into_inner())
+                .map_err(|_| err_msg!(Busy, "Resource handle in use"))
+        })
     }
 
     pub async fn borrow(&self, handle: K) -> Result<TryMutexGuard<V>, Error> {
@@ -105,7 +99,7 @@ where
             .ok_or_else(|| err_msg!("Invalid resource handle"))?
             .1
             .try_lock_arc()
-            .map_err(|_| err_msg!(Busy, "Resource handle in use"))
+            .ok_or_else(|| err_msg!(Busy, "Resource handle in use"))
     }
 
     pub async fn remove_all(&self, store: StoreHandle) -> Result<(), Error> {
@@ -502,8 +496,13 @@ pub extern "C" fn askar_scan_free(handle: ScanHandle) -> ErrorCode {
     catch_err! {
         trace!("Close scan");
         spawn_ok(async move {
-            FFI_SCANS.remove(handle).await.ok();
-            info!("Closed scan {}", handle);
+            // the Scan may have been removed due to the Store being closed
+            if let Some(scan) = FFI_SCANS.remove(handle).await {
+                scan.ok();
+                info!("Closed scan {}", handle);
+            } else {
+                info!("Scan not found for closing: {}", handle);
+            }
         });
         Ok(ErrorCode::Success)
     }
@@ -996,15 +995,20 @@ pub extern "C" fn askar_session_close(
         });
         spawn_ok(async move {
             let result = async {
-                let session = FFI_SESSIONS.remove(handle).await?;
-                if commit == 0 {
-                    // not necessary - rollback is automatic for txn,
-                    // and for regular session there is no action to perform
-                    // > session.rollback().await?;
+                // the Session may have been removed due to the Store being closed
+                if let Some(session) = FFI_SESSIONS.remove(handle).await {
+                    let session = session?;
+                    if commit == 0 {
+                        // not necessary - rollback is automatic for txn,
+                        // and for regular session there is no action to perform
+                        // > session.rollback().await?;
+                    } else {
+                        session.commit().await?;
+                    }
+                    info!("Closed session {}", handle);
                 } else {
-                    session.commit().await?;
+                    info!("Session not found for closing: {}", handle);
                 }
-                info!("Closed session {}", handle);
                 Ok(())
             }.await;
             if let Some(cb) = cb {
